/**
  *  \file client/dialogs/goaldialog.cc
  *  \brief Build Goals Dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/goaldialog.h"
#include "client/widgets/goalinput.h"
#include "client/widgets/helpwidget.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "game/tables.h"

/** Constructor. Builds the complete window.
    \param withUnchange Allow setting to "leave unchanged" */
WBuildGoalsDialog::WBuildGoalsDialog(bool withUnchange)
    : UIWindow(_("Auto-Build Goals"), UIWindow::BLUE_WINDOW, COLOR_GRAY)
{
    createDialog(withUnchange);
}

WBuildGoalsDialog::~WBuildGoalsDialog()
{ }


/** Create one input widget. Input widgets consist of a UIFocusGroup containing
    the actual input widget (UIDecimalSelector), as well as buttons to control it.
    A pointer to the UIDecimalSelector will be stored in input_widgets[col][line].
    \param withUnchange Allow "leave unchanged"
    \param col Column (0=goal, 1=speed)
    \param line Line (structure type)
    \return reference to the whole widget group */
UIBaseWidget&
WBuildGoalsDialog::createInputWidget(bool withUnchange, int col, int line)
{
    const int widget_id   = 1000 + 2*line + col;
    const int lower_bound = withUnchange ? -1 : 0;
    const int upper_bound = col==0 ? 1000 : 100;

    UIFocusGroup& g = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 0, 2));
    UIButton& minus = h.add(new UIButton("-", '-'));
    UIButton& plus  = h.add(new UIButton("+", '+'));
    UIDecimalSelector& ds = h.add(new WGoalInput(widget_id, lower_bound, upper_bound, 10));

    g.add(minus);
    g.add(ds);
    g.add(plus);

    ds.setValue(lower_bound);
    ds.setFocusedWidget();

    input_widgets[col][line] = &ds;

    return g;
}

/** Create dialog. */
void
WBuildGoalsDialog::createDialog(bool withUnchange)
{
    // UIWindow [UIVBoxLayout]
    //   UIGroup [UIGridLayout]
    //     'Structure' 'Goal' 'Speed'
    //     'Mines'     Edit   Edit
    //     'Factories' Edit   Edit
    //     'Defense'   Edit   Edit
    //     'SBD'       Edit   Edit
    //   UIGroup [UIHBoxLayout]
    //     UIButton 'OK'
    //     UIButton 'Cancel'
    //    [UIButton Alt-D]
    //     UISpacer
    //     UIButton 'Help'
    //
    // Each Edit is:
    //     UIFocusGroup [UIHBoxLayout]
    //       UIButton '-'
    //       UIDecimalSelector
    //       UIButton '+'
    setLayout(UIVBoxLayout::instance5);

    UIGroup& group_grid    = h.add(new UIGroup(h.add(new UIGridLayout(3, 0))));
    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    group_grid.add(h.add(new UIStaticText(_("Structure"), UIStaticText::StaticTitle)));
    group_grid.add(h.add(new UIStaticText(_("Goal"),      UIStaticText::StaticTitle)));
    group_grid.add(h.add(new UIStaticText(_("Speed"),     UIStaticText::StaticTitle)));

    for (int i = 0; i < 4; ++i) {
        group_grid.add(h.add(new UIStaticText(getPlanetaryStructureName(i), UIStaticText::StaticTitle)));
        group_grid.add(createInputWidget(withUnchange, 0, i));
        group_grid.add(createInputWidget(withUnchange, 1, i));
    }

    group_buttons.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    if (withUnchange) {
        group_buttons.add(h.add(new UIButton(_("Alt-D - Don't change"), ss_Alt + 'd')));
    }
    group_buttons.add(h.add(new UISpacer()));
    group_buttons.add(h.add(new UIButton(_("Help"), 'h')).addStop());

    input_widgets[0][0]->setFocusedWidget();

    add(group_grid);
    add(group_buttons);
    add(h.add(new WHelpWidget("pcc2:autobuild")));
    pack();
}

/** Event handler. Handles focus hot-keys. */
bool
WBuildGoalsDialog::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* what do we want to do? */
        int delta;
        if (e.key.code == SDLK_TAB)
            delta = 1;
        else if (e.key.code == SDLK_TAB + ss_Shift)
            delta = -1;
        else if (e.key.code == SDLK_DOWN)
            delta = 2;
        else if (e.key.code == SDLK_UP)
            delta = -2;
        else
            delta = 0;

        /* perform focus change */
        if (delta != 0) {
            UIBaseWidget* w = getFocusedWidget();
            if (w != 0) {
                int new_id = w->getId() + delta;
                if (new_id < 1000)
                    new_id += 8;
                else if (new_id >= 1008)
                    new_id -= 8;

                if (UIBaseWidget* new_widget = getWidgetById(new_id))
                    new_widget->setFocusedWidget();
            }
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

/** Initialize with data form a planet. */
void
WBuildGoalsDialog::initData(GPlanet& planet)
{
    for (int i = 0; i < 4; ++i) {
        input_widgets[0][i]->setValue(planet.getAutobuildGoal(TPlanetaryStructure(i)));
        input_widgets[1][i]->setValue(planet.getAutobuildSpeed(TPlanetaryStructure(i)));
    }
}

/** Write back data to a planet. */
void
WBuildGoalsDialog::writeBack(GPlanet& planet)
{
    for (int i = 0; i < 4; ++i) {
        planet.setAutobuildGoal(TPlanetaryStructure(i), input_widgets[0][i]->getValue());
        planet.setAutobuildSpeed(TPlanetaryStructure(i), input_widgets[1][i]->getValue());
    }
}

/** Place focus on the input for a particular structure. */
void
WBuildGoalsDialog::setFocusToStructure(TPlanetaryStructure ps)
{
    input_widgets[0][ps]->setFocusedWidget();
}

/** Get input for a particular goal. */
int
WBuildGoalsDialog::getGoal(TPlanetaryStructure ps) const
{
    return input_widgets[0][ps]->getValue();
}

/** Get input for a particular speed. */
int
WBuildGoalsDialog::getSpeed(TPlanetaryStructure ps) const
{
    return input_widgets[1][ps]->getValue();
}
