/**
  *  \file client/markerconfig.cc
  *  \brief Marker Configuration
  *
  *  Users can define a set of canned markers for quick creation.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/markerconfig.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "client/marks.h"
#include "game/drawing.h"
#include "game/pref.h"

static ConfigMarkerOption marker0(getUserPreferences(), "Chart.Marker0", "2,9,");
static ConfigMarkerOption marker1(getUserPreferences(), "Chart.Marker1", "0,9,");
static ConfigMarkerOption marker2(getUserPreferences(), "Chart.Marker2", "1,9,");
static ConfigMarkerOption marker3(getUserPreferences(), "Chart.Marker3", "2,9,");
static ConfigMarkerOption marker4(getUserPreferences(), "Chart.Marker4", "3,9,");
static ConfigMarkerOption marker5(getUserPreferences(), "Chart.Marker5", "4,9,");
static ConfigMarkerOption marker6(getUserPreferences(), "Chart.Marker6", "5,9,");
static ConfigMarkerOption marker7(getUserPreferences(), "Chart.Marker7", "6,9,");
static ConfigMarkerOption marker8(getUserPreferences(), "Chart.Marker8", "2,10,");
static ConfigMarkerOption marker9(getUserPreferences(), "Chart.Marker9", "0,10,");

/** User's canned markers. */
static ConfigMarkerOption*const canned_markers[10] = {
    &marker0, &marker1, &marker2, &marker3, &marker4,
    &marker5, &marker6, &marker7, &marker8, &marker9,
};

/** Constructor.
    \param container Configuration file
    \param name      Name of configuration file option
    \param deflt     Default value */
ConfigMarkerOption::ConfigMarkerOption(Config& container, const char* name, const char* deflt)
    : ConfigOption(container, name),
      type(0),
      color(1),
      note()
{
    set(deflt);
}

void
ConfigMarkerOption::set(string_t value)
{
    string_t typeStr;
    string_t colorStr;
    int typeVal;
    int colorVal;
    if (strSplit(value, typeStr, value, ",")
        && strSplit(value, colorStr, value, ",")
        && stringValue(typeStr, typeVal)
        && stringValue(colorStr, colorVal)
        && typeVal >= 0
        && typeVal < NUM_USER_MARKERS
        && colorVal >= 0
        && colorVal <= GDrawing::NUM_USER_COLORS)
    {
        type  = typeVal;
        color = colorVal;
        note  = strTrim(value);
    }
}

string_t
ConfigMarkerOption::toString() const
{
    return format("%d,%d,%s", type, color, note);
}

/** Get color index.
    \return index, [0,NUM_USER_COLORS] */
uint8_t
ConfigMarkerOption::getColorIndex() const
{
    return color;
}

/** Set color index.
    \param color index, [0,NUM_USER_COLORS] */
void
ConfigMarkerOption::setColorIndex(uint8_t color)
{
    this->color = color;
}

/** Get marker type.
    \return type, [0,NUM_USER_MARKERS) */
uint8_t
ConfigMarkerOption::getType() const
{
    return type;
}

/** Set marker type.
    \param type type, [0,NUM_USER_MARKERS) */
void
ConfigMarkerOption::setType(uint8_t type)
{
    this->type = type;
}

/** Get note. */
string_t
ConfigMarkerOption::getNote() const
{
    return note;
}

/** Set note.
    \param note New note */
void
ConfigMarkerOption::setNote(string_t note)
{
    this->note = note;
}

/** Get canned marker option. This option can be read or modified.
    \param index Marker index, [0,NUM_CANNED_MARKERS) */
ConfigMarkerOption&
ConfigMarkerOption::getCannedMarker(int index)
{
    return *canned_markers[index];
}
