/**
  *  \file client/wedit.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <algorithm>
#include "client/wedit.h"
#include "cpluslib/assert.h"
#include "util/recode.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "util/console.h"
#include "cpluslib/utf8.h"

/** Construct editor.

    Note that this initializes the line length/count restrictions to
    rather small values, to reflect the fact that this editor does not
    yet support scrolling and hence must fit on a screen. */
WEditor::WEditor(int id)
    : UIWidget(id), max_line_length(50),
      min_line(0), max_line(50), max_storage(50),
      cursor_x(0), cursor_y(0)
{ }

WEditor::~WEditor()
{ }

/** Set maximum line length. The editor will avoid that users create
    lines longer than this. */
void
WEditor::setMaxLineLength(col_t len)
{
    max_line_length = len;
}

/** Set maximum number of stored lines. The editor will store at most
    this many lines. This may be more than the user is allowed to
    see. */
void
WEditor::setMaxStoredLines(line_t lines)
{
    max_storage = lines;
    if (max_storage < content.size())
        content.erase(content.begin() + max_storage, content.end());
}

/** Set line restrictions for user.
    \param min Minimum line the user is allowed to edit (0=first), i.e.
               number of uneditable lines at front
    \param max One-past maximum line the user is allowed to edit */
void
WEditor::setUserLineRestrictions(line_t min, line_t max)
{
    ASSERT(min < max);
    min_line = min;
    max_line = max;
    if (cursor_y < min_line)
        cursor_y = min_line;
    if (cursor_y >= max_line)
        cursor_y = max_line-1;
    drawWidget();
}

/** Insert blank line.
    \param line Line that will be shifted down and replaced by blank line */
void
WEditor::insertLine(line_t line)
{
    if (line >= content.size())
        return;
    if (content.size() < max_storage)
        content.push_back(string_t());
    for (line_t i = content.size()-1; i > line; --i)
        content[i] = content[i-1];
    content[line] = string_t();
    drawWidget();
}

/** Delete lines.
    \param line First line to delete
    \param count Number of lines to delete */
void
WEditor::deleteLine(line_t line, line_t count)
{
    /* we have situation:
         [0, line) [line, line+count) [line+count, size)
       and want
         [0, line) [line+count, size) */
    if (count == 0) {
        /* nothing */
    } else if (content.size() < line + count) {
        if (content.size() > line)
            content.erase(content.begin() + line, content.end());
    } else {
        vec_t::iterator dest = content.begin() + line,
            src = dest + count;
        while (src != content.end()) {
            *dest = *src;
            ++src, ++dest;
        }
        content.erase(dest, content.end());
    }
    drawWidget();
}

/** Get line.
    \param line Line to read out (0=first) */
string_t
WEditor::getLine(line_t line) const
{
    if (line < content.size())
        return content[line];
    else
        return string_t();
}

/** Set line. Replace a line with new text.
    \param line Line to replace (0=first)
    \param text Text to put in line */
void
WEditor::setLine(line_t line, string_t text)
{
    if (line >= max_storage) {
        /* out of range */
    } else if (line >= content.size()) {
        if (!text.empty()) {
            content.insert(content.end(), line - content.size(), string_t());
            content.push_back(text);
            drawWidget();
        }
    } else {
        if (text != content[line]) {
            content[line] = text;
            drawWidget();
        }
    }
}

/** Get contents of editor as string. Concatenates all lines with
    "\n" between, and returns the result. */
string_t
WEditor::getContentAsString() const
{
    return getContentAsString(0, max_line);
}

/** Get contents of editor as string. Concatenates all lines between
    begin and end, with "\n" between, and returns the result.
    \param begin First line to include (0=first)
    \param end One-past last line */
string_t
WEditor::getContentAsString(line_t begin, line_t end) const
{
    string_t rv;
    if (end > content.size())
        end = content.size();
    for (line_t i = begin; i < end; ++i) {
        rv += content[i];
        rv += '\n';
    }
    return rv;
}

static bool
canBreakLineAt(uint32_t c)
{
    return c == ' ' || c == '-' || c == '/';
}

void
WEditor::splitLineNoRedraw(line_t line, col_t col)
{
    if (line < content.size()) {
        insertLine(line + 1);
        if (lengthUtf8(content[line]) > col) {
            setLine(line + 1, strRTrim(substrUtf8(content[line], col, string_t::npos)));
            setLine(line,     strRTrim(substrUtf8(content[line], 0, col)));
        }
    }

    if (cursor_y == line && cursor_x >= col) {
        ++cursor_y;
        cursor_x -= col;
    }
}

void
WEditor::wrapLineNoRedraw(line_t line, col_t length)
{
    if (length <= 0)
        return;                 // avoid stupid case

    while (line < content.size() && lengthUtf8(content[line]) > length) {
        col_t split_at = length;
        for (col_t i = length; i > 1; --i) {
            if (canBreakLineAt(charAtUtf8(content[line], i-1))) {
                split_at = i;
                break;
            }
        }

        splitLineNoRedraw(line, split_at);
        ++line;
    }
}

/** Insert text at the current cursor position. */
void
WEditor::insertNoRedraw(string_t c)
{
    string_t line = getLine(cursor_y);
    if (lengthUtf8(line) < cursor_x)
        line.append(cursor_x - lengthUtf8(line), ' ');

    string_t new_line = substrUtf8(line, 0, cursor_x) + c;
    if (lengthUtf8(new_line) < max_line_length)
        new_line.append(substrUtf8(line, cursor_x, max_line_length - lengthUtf8(new_line)));
    setLine(cursor_y, strRTrim(new_line));

    cursor_x += lengthUtf8(c);
    wrapLineNoRedraw(cursor_y, max_line_length);

    /* if wrapping didn't get the cursor back in, we were at the
       rightmost column and typed a space. Move to next line. */
    if (cursor_x > max_line_length) {
        if (cursor_y < max_line-1) {
            ++cursor_y;
            insertLine(cursor_y);
            cursor_x = 0;
        } else {
            cursor_x = max_line_length;
        }
    }

    if (cursor_y >= max_line) {
        cursor_y = max_line-1;
        cursor_x = lengthUtf8(getLine(cursor_y));
    }
}

void
WEditor::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();
    ctx.useFont(*fonts[FONT_FIXED]);

    ctx.setColor(UIColor::tc_White);
    for (line_t i = 0; i < max_line; ++i) {
        // FIXME: maybe show unselectable lines in different color?
        outTextF(ctx, getExtent().x,
                 getExtent().y + i * font_heights[FONT_FIXED],
                 getExtent().w, getLine(i));
    }

    ctx.setColor(UIColor::tc_Yellow);
    drawVLine(ctx,
              getExtent().x + cursor_x * font_ems[FONT_FIXED],
              getExtent().y + cursor_y * font_heights[FONT_FIXED],
              getExtent().y + (cursor_y+1) * font_heights[FONT_FIXED]);

    can.update(getExtent());
}

void
WEditor::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size =
        font_sizes[FONT_FIXED].scaledBy(max_line_length, max_line).movedBy(1, 0);
}

void
WEditor::doDelete(bool forward)
{
    if (!forward) {
        if (cursor_x == 0) {
            if (cursor_y <= min_line)
                return;
            --cursor_y;
            cursor_x = lengthUtf8(getLine(cursor_y));
        } else {
            --cursor_x;
        }
    }

    string_t line = getLine(cursor_y);
    if (cursor_x >= lengthUtf8(line)) {
        setLine(cursor_y, line + getLine(cursor_y + 1));
        deleteLine(cursor_y + 1);
        wrapLineNoRedraw(cursor_y, max_line_length);
    } else {
        line = substrUtf8(line, 0, cursor_x) + substrUtf8(line, cursor_x+1, string_t::npos);
        setLine(cursor_y, line);
    }
    drawWidget();
}

bool
WEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_UP:
            if (cursor_y > min_line) {
                setSelectedWidget();
                --cursor_y;
                drawWidget();
            }
            return true;
         case SDLK_DOWN:
            if (cursor_y+1 < max_line) {
                setSelectedWidget();
                ++cursor_y;
                drawWidget();
            }
            return true;
         case SDLK_LEFT:
            if (cursor_x) {
                setSelectedWidget();
                --cursor_x;
                drawWidget();
            }
            return true;
         case SDLK_RIGHT:
            if (cursor_x < max_line_length) {
                setSelectedWidget();
                ++cursor_x;
                drawWidget();
            }
            return true;
         case SDLK_HOME:
            setSelectedWidget();
            cursor_x = 0;
            drawWidget();
            return true;
         case SDLK_END:
            setSelectedWidget();
            cursor_x = lengthUtf8(getLine(cursor_y));
            drawWidget();
            return true;
         case SDLK_RETURN:
            setSelectedWidget();
            if (cursor_y+1 < max_line)
                splitLineNoRedraw(cursor_y, cursor_x);
            drawWidget();
            return true;
         case SDLK_DELETE:
            setSelectedWidget();
            doDelete(true);
            return true;
         case SDLK_BACKSPACE:
            setSelectedWidget();
            doDelete(false);
            return true;
         case ss_Ctrl + 'y':
            setSelectedWidget();
            deleteLine(cursor_y, 1);
            drawWidget();
            return true;
         default:
            if (e.key.unicode >= 32 && e.key.unicode != 127 && (e.key.code & ss_AllMod) == 0) {
                if (isReliableGameCharacter(e.key.unicode) && uint8_t(getGameCharacterSet().getCharacter(e.key.unicode)) < 256-13) {
                    string_t tmp;
                    appendUtf8(tmp, e.key.unicode);
                    setSelectedWidget();
                    insertNoRedraw(tmp);
                    drawWidget();
                    return true;
                }
            }
        }
    } else if (e.type == ev_Mouse && e.mouse.b && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        unsigned int nx = (e.mouse.x - getExtent().x) / font_ems[FONT_FIXED];
        unsigned int ny = (e.mouse.y - getExtent().y) / font_heights[FONT_FIXED];
        if (ny >= min_line && ny < max_line) {
            if (nx != cursor_x || ny != cursor_y) {
                cursor_x = nx;
                cursor_y = ny;
                drawWidget();
            }
        }
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Insert string at cursor. The string will be inserted verbatim. It
    can contain "\n" to force line breaks. */
void
WEditor::insertString(string_t s)
{
    string_t::size_type n, p = 0;
    while ((n = s.find('\n', p)) != s.npos) {
        insertNoRedraw(s.substr(p, n - p));
        if (cursor_y+1 < max_line)
            splitLineNoRedraw(cursor_y, cursor_x);
        p = n+1;
    }
    insertNoRedraw(s.substr(p));
    drawWidget();
}
