/**
  *  \file client/widgets/msgdisplay.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/msgdisplay.h"
#include "client/dlg-subject.h"
#include "client/screens.h"     // doSendMessage
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/global.h"        // getDisplayedTurn
#include "game/msg.h"
#include "gfx/gfxcontext.h"
#include "io/streams.h"
#include "io/textfile.h"        // TextFile
#include "ui/dialogs/filedlg.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/recode.h"        // getGameCharacterSet
#include "io/glob.h"
#include "ui/widgets/inputline.h"
#include "client/widgets/setselect.h"
#include "client/dialogs/msgreceiver.h"

static string_t message_search_string;

/** Find non-filtered message.
    \param mbox Mailbox to search
    \param index Start search here
    \param dir Direction */
static int
skipFiltered(GMessagebox& mbox, int index, int dir)
{
    const GMessageFilter& filter = getMessageFilter();
    int i = index;
    while (i >= 0 && i < mbox.getCount()) {
        if (!filter.isFiltered(mbox.getHeading(i))) {
            return i;
        }
        i += dir;
    }
    return index;
}

/** \class WMessageDisplay
    \brief Widget displaying contents of a mailbox

    This widget displays one message of a mailbox.
    It automatically updates when the user chooses another message.
    It provides reaction to various keypresses.
    It generates IntUserInterfaceBinding callbacks. */
WMessageDisplay::WMessageDisplay(int id, GMessagebox& mbox)
    : UIWidget(id), mbox(mbox)
{
    conn = observe(mbox.sig_sel_changed);
}

WMessageDisplay::~WMessageDisplay()
{ }

void
WMessageDisplay::searchFromMessage(int nr)
{
    string_t substr = strUCase(message_search_string);
    while (nr < mbox.getCount()) {
        if (strUCase(mbox.getText(nr)).find(substr) != string_t::npos) {
            mbox.setCurrent(nr);
            return;
        }
        ++nr;
    }
    messageBox(_("Search text not found."), _("Search in messages"));
}

bool
WMessageDisplay::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_TAB:
            {
                WSubjectList subjlist(1000, mbox, getMessageFilter());
                subjlist.selectMessage(mbox.getCurrent());
                if (subjlist.doStandardDialog())
                    mbox.setCurrent(subjlist.getSelectedMessage());
            }
            return true;

         case SDLK_PAGEUP:
         case SDLK_UP:
         case vk_WheelUp:
         case '-':
            doPrev(false, e.key.arg);
            return true;

         case ss_Shift + SDLK_PAGEUP:
         case ss_Shift + SDLK_UP:
         case ss_Shift + vk_WheelUp:
            doPrev(true, e.key.arg);
            return true;

         case SDLK_PAGEDOWN:
         case SDLK_DOWN:
         case vk_WheelDown:
         case '+':
            doNext(false, e.key.arg);
            return true;

         case ss_Shift + SDLK_PAGEDOWN:
         case ss_Shift + SDLK_DOWN:
         case ss_Shift + vk_WheelDown:
            doNext(true, e.key.arg);
            return true;

         case ss_Ctrl + SDLK_HOME:
         case ss_Ctrl + SDLK_PAGEUP:
         case '<':
         case '=':
            if (e.key.arg != 0) {
                mbox.setCurrent(e.key.arg-1);
            } else {
                mbox.setCurrent(skipFiltered(mbox, 0, 1));
            }
            return true;

         case ss_Ctrl + SDLK_END:
         case ss_Ctrl + SDLK_PAGEDOWN:
         case '>':
            if (e.key.arg != 0) {
                mbox.setCurrent(mbox.getCount()-1);
            } else {
                mbox.setCurrent(skipFiltered(mbox, mbox.getCount()-1, -1));
            }
            return true;

         case 's':
         case '/':
         case SDLK_F7:
            doSearch();
            return true;

         case 'n':
         case ss_Shift + SDLK_F7:
            doSearchNext();
            return true;

         case 'f':
            doForwardMessage();
            return true;

         case 'w':
            doWriteMessage(false);
            return true;

         case ss_Ctrl + 'w':
            doWriteMessage(true);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
WMessageDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.min_size = info.pref_size = font_sizes[FONT_FIXED].scaledBy(41, 22);
}

void
WMessageDisplay::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_FIXED]);
    ctx.setColor(UIColor::tc_White);

    drawBackground(ctx, getExtent());
    outTextFormattedF(ctx, getExtent().x, getExtent().y,
                      getExtent().w, getExtent().h / font_heights[FONT_FIXED],
                      mbox.getText(mbox.getCurrent()));

    can.update(getExtent());
}

/** Go to next message.
    \param killed also accept killed messages
    \param count number of messages to advance */
void
WMessageDisplay::doNext(bool killed, int count)
{
    const GMessageFilter& filter = getMessageFilter();
    int n = mbox.getCurrent();
    if (!killed) {
        killed = (count != 0 || filter.isFiltered(mbox.getHeading(n)));
    }
    if (count <= 0) {
        count = 1;
    }

    // Find next message
    bool found = false;
    while (!found && ++n < mbox.getCount()) {
        if (killed || !filter.isFiltered(mbox.getHeading(n))) {
            if (--count == 0) {
                found = true;
            }
        }
    }

    if (found) {
        mbox.setCurrent(n);
    }
}

/** Go to previous message.
    \param killed also accept killed messages
    \param count number of messages to advance */
void
WMessageDisplay::doPrev(bool killed, int count)
{
    GMessageFilter& filter = getMessageFilter();
    int n = mbox.getCurrent();
    if (!killed) {
        killed = (count != 0 || filter.isFiltered(mbox.getHeading(n)));
    }
    if (count <= 0) {
        count = 1;
    }

    // Find previous message
    bool found = false;
    while (!found && --n >= 0) {
        if (killed || !filter.isFiltered(mbox.getHeading(n))) {
            if (--count == 0) {
                found = true;
            }
        }
    }

    if (found) {
        mbox.setCurrent(n);
    }
}

/** Search for a string. */
void
WMessageDisplay::doSearch()
{
    if (UIInputLine::run(message_search_string, 40, _("Search in messages"), _("Search for:"), 20, 0)
        && message_search_string.length())
    {
        searchFromMessage(0);
    }
}

/** Continue a search operation. */
void
WMessageDisplay::doSearchNext()
{
    if (message_search_string.length()) {
        searchFromMessage(mbox.getCurrent()+1);
    } else {
        doSearch();
    }
}

/** Forward this message to somebody. */
void
WMessageDisplay::doForwardMessage()
{
    WPlayerSetSelector sel(0, NUM_OWNERS, 15*font_ems[FONT_NORMAL]);
    if (chooseMessageReceiver(sel, _("Forward Message"), "pcc2:msgin")) {
        doSendMessage("--- Forwarded Message ---\n"
                      + mbox.getText(mbox.getCurrent())
                      + "\n--- End Forwarded Message ---",
                      sel.getSelectedPlayers());
    }        
}

/** Write message to file.
    \param all Write all messages, not just this one */
void
WMessageDisplay::doWriteMessage(bool all)
{
    // Ask for file name
    const char* heading = all ? _("Save All Messages") : _("Save this Message");
    UIFileChooserDialog select(heading,
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("txt"));
    ui_root->centerWidget(select);
    if (!ui_root->runModal(select))
        return;

    // Create file
    Ptr<Stream> s = Stream::openFileNT(select.getResultFile(), Stream::C_WRITE);
    if (!s) {
        try {
            s = Stream::openFile(select.getResultFile(), Stream::C_CREATE);
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to create file %s: %s"), e.getFileName(), e.what()), heading);
            return;
        }
        catch (std::exception& e) {
            messageBox(format(_("Unable to create file %s: %s"), select.getResultFile(), e.what()), heading);
            return;
        }
    }
    s->seek(s->getSize());

    // Write message. Use game character set, because programs like PCC 1.x / mgrep
    // use these files and assume game character set. The message will not contain
    // anything else, anyway.
    // FIXME: the turn number may not be correct?
    TextFile tf(*s);
    tf.setCharacterSet(getGameCharacterSet());
    tf.writeLine(format("=== Turn %d ===", getDisplayedTurn().getTurnNumber()));
    if (all) {
        tf.writeLine(format("   %d message(s)", mbox.getCount()));
        for (int i = 0, n = mbox.getCount(); i < n; ++i) {
            tf.writeLine(format("--- Message %d ---", i+1));
            tf.writeLine(mbox.getText(i));
        }
    } else {
        int i = mbox.getCurrent();
        tf.writeLine(format("--- Message %d ---", i+1));
        tf.writeLine(mbox.getText(i));
    }
    tf.sync();
}
