/**
  *  \file game/ufo.cc
  *  \brief Ufo and Wormhole handling
  *
  *  Ufo loading mechanism: the most important thing Ufos are used for
  *  wormholes. We handle these specially. Other than that, UFO.HST
  *  Ufos and PHost's General Objects (GOs) all end up here. We ought
  *  to identify an object by its Id/Type-Code combination as we
  *  cannot assume that all add-ons that generate GOs coordinate on an
  *  Id range. This isn't yet implemented, though, making PCC2 behave
  *  identical to PCC 1.x. In any case, Ufos with an Id in the range
  *  of 1..1000 need not have the same type code, and multiple Ufos
  *  with the same Id/Type are not supported either.
  *
  *  For Ufos and GOs, merging is trivial. We assume both data sources
  *  contain equivalent information; the last seen instance survives.
  *
  *  For wormholes, we have three possible sources:
  *  - the Ufo from KORE.DAT
  *  - the UTIL.DAT entry
  *  - the WORMHOLE.TXT file
  *  Unfortunately, there is no 1:1 mapping between wormhole Ids and
  *  Ufo Ids (each WH consumes two WH Id slots, but whereas a
  *  bidirectional WH consumes two Ufo slots, an unidirectional one
  *  consumes only one). We therefore queue all UTIL.DAT wormholes
  *  first, and merge them later upon postprocess() time. In the
  *  single player case, we can simply match the Ufo and wormhole
  *  sequences: if KORE.DAT contains Ufos 51,53,54, and UTIL.DAT
  *  contains Ufos 0,6,7, we can therefore derive the mapping. It's a
  *  little harder if we have multiple players with different
  *  registration status.
  *
  *  Finally, WORMHOLE.TXT can be used to "fill in the blanks", mainly
  *  intended to be used in games where wormholes are static and known
  *  to everyone in the game.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/ufo.h"
#include "game/struct.h"
#include "game/utildata.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "cpluslib/misc.h"
#include "util/nls.h"
#include "util/console.h"
#include "game/tables.h"
#include "game/formula.h"
#include "game/config.h"
#include "game/univ.h"
#include "util/recode.h"

/** Estimate movement of a wormhole.
    \param pos_now [in] Current position (X or Y)
    \param pos_old [in] Old position (X or Y)
    \param vec     [in/out] Movement per turn, guess
    \param time    [in] Time that has passed between pos_old and pos_now */
static void
estimateMovement(int pos_now, int pos_old, int& vec, int time)
{
    if (config.WrmDisplacement() == 0) {
        /* We know wormholes don't move (other than possible Brownian movement
           through WrmRandDisplacement) */
        vec = 0;
    } else {
        /* Wormhole moves by n*WrmDisplacement, plus r*WrmRandDisplacement,
           where n, r are from [-1,+1], n is deterministic, r is random.
           We want to know n.

                           |WrmRandDisplacement|
           |WrmRandDisplacement|           |WrmRandDisplacement|
                     |WrmDisplacement|WrmDisplacement| */
        // FIXME: adjust /dif/ for wraparound
        int dif = pos_now - pos_old;
        if (dif > 0 && dif > time * config.WrmRandDisplacement()) {
            /* We moved to the right by more than WrmRandDisplacement,
               so that must be n>0 */
            vec = config.WrmDisplacement();
        } else if (dif < 0 && dif < -time * config.WrmRandDisplacement()) {
            /* We moved to the left by more then WrmRandDisplacement,
               so that must be n<0 */
            vec = -config.WrmDisplacement();
        } else if (std::abs(dif) <= time * config.WrmRandDisplacement()
                   && ((config.WrmDisplacement() > 2*config.WrmRandDisplacement())
                       || std::abs(dif) < time * (config.WrmDisplacement() - config.WrmRandDisplacement())))
        {
            /* We moved within the bounds of WrmRandDisplacement, and
               that position cannot be achieved by WrmDisplacement. */
            vec = 0;
        } else {
            /* No usable information derivable */
        }
    }
}

/********************************** GUfo *********************************/

/** Create blank Ufo. This will initialize the Ufo to all-zero.
    \param id Id number */
GUfo::GUfo(int id)
    : id(id),
      vga_color(0),
      pos(0, 0),
      warp(0),
      heading(-1),
      planet_range(0), ship_range(0),
      radius(0),
      type_code(0),
      name(),
      info1(), info2(),
      real_id(0),
      turn_last_seen(0),
      pos_last_seen(0, 0),
      movement_vector(0, 0),
      flags(0),
      other_end(0)
{ }

/** Copy constructor. Makes a clone of the specified Ufo. */
GUfo::GUfo(const GUfo& other)
    : GCircularObject(),
      id(other.id),
      vga_color(other.vga_color),
      pos(other.pos),
      warp(other.warp),
      heading(other.heading),
      planet_range(other.planet_range),
      ship_range(other.ship_range),
      radius(other.radius),
      type_code(other.type_code),
      name(other.name),
      info1(other.info1),
      info2(other.info2),
      real_id(other.real_id),
      turn_last_seen(other.turn_last_seen),
      pos_last_seen(other.pos_last_seen),
      movement_vector(other.movement_vector),
      flags(other.flags),
      other_end(0)
{ }

/** Destructor. */
GUfo::~GUfo()
{
    disconnect();
}

/** Copy operator. Makes this Ufo a copy of \c rhs. */
GUfo&
GUfo::operator=(const GUfo& rhs)
{
    disconnect();
    id              = rhs.id;
    vga_color       = rhs.vga_color;
    pos             = rhs.pos;
    warp            = rhs.warp;
    heading         = rhs.heading;
    planet_range    = rhs.planet_range;
    ship_range      = rhs.ship_range;
    radius          = rhs.radius;
    type_code       = rhs.type_code;
    name            = rhs.name;
    info1           = rhs.info1;
    info2           = rhs.info2;
    real_id         = rhs.real_id;
    turn_last_seen  = rhs.turn_last_seen;
    pos_last_seen   = rhs.pos_last_seen;
    movement_vector = rhs.movement_vector;
    flags           = rhs.flags;
    return *this;
}

/* GCircularObject interface */
string_t
GUfo::getName(Name which) const
{
    switch (which) {
     case PlainName:
        return name;
     case LongName:
     case DetailedName:
        return format("Ufo #%d: %s", id, name);
    }
    return string_t();
}

int
GUfo::getId() const
{
    return id;
}

int
GUfo::getOwner() const
{
    return 0;
}

bool
GUfo::isExistant() const
{
    return vga_color != 0;
}

GPoint
GUfo::getPos() const
{
    return pos;
}

int
GUfo::getRadius() const
{
    return radius;
}

int32_t
GUfo::getRadiusSquared() const
{
    return int32_t(radius) * radius;
}

/* Loader interface */

/** Add data from KORE.DAT Ufo entry. */
void
GUfo::addUfoData(const TUfo& data)
{
    copyUfo(data);
    flags |= UfoSeenThisTurn;
}

/** Add data from UTIL.DAT entry. */
void
GUfo::addObjectData(const TUtil33GO& obj)
{
    ASSERT(obj.oid == id);

    /* Accept this data only if we don't yet have current data. If it
       is an object for which we have both a KORE entry and a UTIL
       entry, this means we keep the KORE entry, which has ship_range
       and planet_range fields. */
    if (!isSeenThisTurn()) {
        vga_color    = obj.color;
        pos          = GPoint(obj.x, obj.y);
        warp         = obj.speed;
        heading      = obj.heading;
        planet_range = 0;
        ship_range   = 0;
        radius       = obj.radius;
        name         = convertGameToUtf8(getBasicString(obj.name));
        info1        = convertGameToUtf8(getBasicString(obj.info1));
        info2        = convertGameToUtf8(getBasicString(obj.info2));
        type_code    = obj.type;
    }

    flags |= UfoSeenThisTurn;
}

/** Add data from history. */
void
GUfo::addHistoryData(const TDbUfo& data)
{
    /* Do not overwrite the Ufo data when we saw this thing this turn */
    ASSERT(data.id == id);
    if (!isSeenThisTurn())
        copyUfo(data.ufo);
    real_id         = data.real_id;
    turn_last_seen  = data.turn_last_seen;
    pos_last_seen   = GPoint(data.x_last_seen, data.y_last_seen);
    movement_vector = GPoint(data.x_move, data.y_move);
    flags          |= UfoStoredInDatabase;
}

/** Merge wormhole data.
    \param data Data from util.dat
    \param all Merge all data. false=no, this is the Ufo we got from host; true=yes, this
               is an all-new Ufo */
void
GUfo::addWormholeData(const TUtil14Wormhole& data, bool all, int turn)
{
    // ccmain.pas::ConsumeWormhole
    if (all) {
        /* Making a new wormhole Ufo, so fill in header */
        name  = format(_("Wormhole #%d"), data.worm_id);

        /* Mass/directionality */
        const char* fmt = (data.bidir == 0
                           ? _("%d kt/Enter only")
                           : data.bidir == 1
                           ? _("%d kt/Bidir.")
                           : _("%d kt"));
        info1 = format(fmt, data.mass);
    } else {
        /* Just updating an old Ufo */
        if (pos.x != data.x || pos.y != data.y)
            console.write(LOG_WARN, format(_("Ufo #%d and wormhole #%d do not match."), id, data.worm_id));
    }

    real_id   = data.worm_id;
    vga_color = 2;              /* green, same which PHost uses */
    type_code = 1;              /* same code as the one used by PHost */
    info2     = getWormholeStabilityName(data.stable_code);
    pos.x     = data.x;
    pos.y     = data.y;
    warp      = 0;
    heading   = -1;

    if (data.mass <= 0) {
        // degenerate case
        planet_range = ship_range = 0;
        radius = 2;
    } else {
        // normal case
        planet_range = ship_range = roundToInt(10 * std::exp(std::log(double(data.mass)) / 3));
        radius = roundToInt(exp(config.WrmEntryPowerX100()*0.01 * std::log(double(data.mass))) / 2);
    }

    flags |= UfoStoredInDatabase | UfoSeenThisTurn;

    if (turn_last_seen > 0 && turn_last_seen < turn) {
        // estimate movement
        estimateMovement(pos.x, pos_last_seen.x, movement_vector.x, turn - turn_last_seen);
        estimateMovement(pos.y, pos_last_seen.y, movement_vector.y, turn - turn_last_seen);
    }
}

/** Disconnect this Ufo from its other end. */
void
GUfo::disconnect()
{
    if (other_end != 0) {
        other_end->other_end = 0;
        other_end = 0;
    }
}

/** Connect this Ufo with another one. Used to establish wormhole end links. */
void
GUfo::connectWith(GUfo& other)
{
    disconnect();
    other.disconnect();
    other_end = &other;
    other.other_end = this;
}

/** Postprocess after loading. */
void
GUfo::postprocess(int turn)
{
    if (!isSeenThisTurn() && turn_last_seen > 0) {
        pos.x = pos_last_seen.x + movement_vector.x * (turn - turn_last_seen);
        pos.y = pos_last_seen.y + movement_vector.y * (turn - turn_last_seen);
    }
    if (isSeenThisTurn() && turn_last_seen < turn) {
        turn_last_seen = turn;
        pos_last_seen = pos;
    }
}

/** Get data to store in history. */
void
GUfo::getHistoryData(TDbUfo& data) const
{
    data.id               = id;

    data.ufo.color        = vga_color;
    storeBasicString(data.ufo.name, convertUtf8ToGame(name));
    storeBasicString(data.ufo.info1, convertUtf8ToGame(info1));
    storeBasicString(data.ufo.info2, convertUtf8ToGame(info2));
    data.ufo.x            = pos.x;
    data.ufo.y            = pos.y;
    data.ufo.warp         = warp;
    data.ufo.heading      = heading;
    data.ufo.planet_range = planet_range;
    data.ufo.ship_range   = ship_range;
    data.ufo.radius       = radius;
    data.ufo.type_code    = type_code;

    data.real_id          = real_id;
    data.turn_last_seen   = turn_last_seen;
    data.x_last_seen      = pos_last_seen.x;
    data.y_last_seen      = pos_last_seen.y;
    data.x_move           = movement_vector.x;
    data.y_move           = movement_vector.y;
}

/** Set whether to keep this Ufo in the history database. */
void
GUfo::setIsStoredInHistory(bool value)
{
    if (value != isStoredInHistory()) {
        flags ^= UfoStoredInDatabase;
        markDirty();
    }
}

int
GUfo::getColorCode() const
{
    return vga_color;
}

/** Get color to use for this Ufo.
    \returns palette index */
uint8
GUfo::getColor() const
{
    static const uint8 vga_to_ours[] = {
        COLOR_BLACK,     COLOR_DARKBLUE,      COLOR_DARKGREEN,    COLOR_DARKCYAN,
        COLOR_DARKRED,   COLOR_DARKMAGENTA,   COLOR_DARKYELLOW,   COLOR_GRAY,
        COLOR_DARK,      COLOR_BRIGHTBLUE,    COLOR_BRIGHTGREEN,  COLOR_BRIGHTCYAN,
        COLOR_BRIGHTRED, COLOR_BRIGHTMAGENTA, COLOR_BRIGHTYELLOW, COLOR_WHITE
    };
    if (vga_color >= 0 && vga_color < int(countof(vga_to_ours)))
        return vga_to_ours[vga_color];
    else
        return COLOR_WHITE;
}

int
GUfo::getSpeed() const
{
    return warp;
}

int
GUfo::getHeading() const
{
    return heading;
}

int
GUfo::getPlanetRange() const
{
    return planet_range;
}

int
GUfo::getShipRange() const
{
    return ship_range;
}

int
GUfo::getTypeCode() const
{
    return type_code;
}

string_t
GUfo::getInfo1() const
{
    return info1;
}

string_t
GUfo::getInfo2() const
{
    return info2;
}

int32
GUfo::getRealId() const
{
    return real_id;
}

GPoint
GUfo::getOtherEndPos() const
{
    if (other_end)
        return other_end->getPos();
    else
        return GPoint();
}

int
GUfo::getLastTurn() const
{
    return turn_last_seen;
}

GPoint
GUfo::getMovementVector() const
{
    return movement_vector;
}

GUfo*
GUfo::getOtherEnd() const
{
    return other_end;
}

/** Copy from Ufo record. */
void
GUfo::copyUfo(const TUfo& data)
{
    vga_color    = data.color;
    name         = convertGameToUtf8(getBasicString(data.name));
    info1        = convertGameToUtf8(getBasicString(data.info1));
    info2        = convertGameToUtf8(getBasicString(data.info2));
    pos          = GPoint(data.x, data.y);
    warp         = data.warp;
    heading      = data.heading;
    planet_range = data.planet_range;
    ship_range   = data.ship_range;
    radius       = data.radius;
    type_code    = data.type_code;
}

/******************************** GUfoType *******************************/

GUfoType::GUfoType(GUniverse& univ)
    : univ(univ),
      ufos(),
      num_ufos(0),
      wormholes()
{ }

GUfoType::~GUfoType()
{ }

/** Add wormhole data from util.dat. */
void
GUfoType::addWormholeData(const TUtil14Wormhole& wh)
{
    wormholes.push_back(wh);
}

/** Add Ufo data from current game file.
    This is a member of GUfoType to permit switching to id/type-based
    addressing somewhere in the future. */
void
GUfoType::addUfoData(int id, const TUfo& ufo)
{
    getUfoById(id).addUfoData(ufo);
}

/** Add General Object data from UTIL.DAT. */
void
GUfoType::addObjectData(const TUtil33GO& obj)
{
    getUfoById(obj.oid).addObjectData(obj);
}

/** Add history data from database game file.
    This is a member of GUfoType to permit switching to id/type-based
    addressing somewhere in the future. */
void
GUfoType::addHistoryData(const TDbUfo& ufo)
{
    getUfoById(ufo.id).addHistoryData(ufo);
}

/** Postprocess Ufos after loading. */
void
GUfoType::postprocess()
{
    /* Convert util.dat wormholes into Ufos */
    if (!wormholes.empty()) {
        /* There is no reproducible 1:1 correspondence between wormhole Ids
           and Ufo Ids. PHost allocates two wormhole slots for each type of
           wormhole, but unidirectional wormholes get just one Ufo slot. Hence,
           we assume that the sequence of Ufos matches the sequence of wormholes
           if there are some Ufos inside the reserved range. If there are no
           Ufos (because the player is using Dosplan), we generate the sequence
           internally. */
        // FIXME: this does not deal with multiple RSTs. They may even have
        // different registration status! Things to do to support multiple RSTs:
        // - sort/uniq wormholes
        // - track sources for Ufos/wormholes

        /* First, find first wormhole Ufo. */
        iterator i = getFirstAt(config.WormholeUFOsStartAt());
        while (i != end() && !i->isSeenThisTurn())
            ++i;

        /* Regular merging */
        while (!wormholes.empty() && i != end() && i->getId() < config.WormholeUFOsStartAt() + 200) {
            /* Merge */
            i->addWormholeData(wormholes.front(), false, univ.getTurnNumber());

            /* Advance to next Ufo; skip those that we didn't see this turn */
            do
                ++i;
            while (i != end() && !i->isSeenThisTurn());

            /* Drop wormhole */
            wormholes.pop_front();
        }

        /* Generating new Ufos */
        while (!wormholes.empty()) {
            getUfoById(wormholes.front().worm_id + config.WormholeUFOsStartAt()).addWormholeData(wormholes.front(), true, univ.getTurnNumber());
            wormholes.pop_front();
        }
    }

    /* Postprocessing. This updates guessed positions. */
    for (iterator i = begin(); i != end(); ++i)
        i->postprocess(univ.getTurnNumber());

    /* Connect wormhole Ufos */
    for (iterator i = getFirstAt(config.WormholeUFOsStartAt());
         i != end() && i->getId() < config.WormholeUFOsStartAt() + 200;
         ++i)
    {
        /* We can connect this Ufo with the next one if it has an even
           real_id, and the next one exists and has a one-higher Id */
        iterator n = i;
        if ((i->getRealId() & 1) == 0 && ++n != end() && n->getRealId() == i->getRealId()+1) {
            i->connectWith(*n);
        }
    }
}

GUfo&
GUfoType::getObjectByIndex(int index)
{
    return getUfoByIndex(index);
}

GUniverse*
GUfoType::getUniverseByIndex(int /*index*/)
{
    return &univ;
}

bool
GUfoType::isValidIndex(int index) const
{
    return index > 0
        && index <= num_ufos;
}

int
GUfoType::getNextIndex(int index) const
{
    if (index < num_ufos)
        return index+1;
    else
        return 0;
}

int
GUfoType::getPreviousIndex(int index) const
{
    if (index == 0)
        return num_ufos;
    else
        return index-1;
}

/** Get first Ufo with an Id of id or higher. */
GUfoType::iterator
GUfoType::getFirstAt(int id)
{
    iterator i = begin();
    while (i != end() && i->getId() < id)
        ++i;
    return i;
}

/** Get Ufo with specified Id. Creates it if it does not exist. */
GUfo&
GUfoType::getUfoById(int id)
{
    iterator i = ufos.begin();
    while (i != ufos.end() && i->getId() < id)
        ++i;

    if (i == ufos.end()) {
        /* does not exist, and has higher Id than we know so far */
        ++num_ufos;
        ufos.push_back(GUfo(id));
        return ufos.back();
    } else if (i->getId() == id) {
        /* does already exist */
        return *i;
    } else {
        /* does not exist, have to insert in the middle */
        ++num_ufos;
        return *ufos.insert(i, GUfo(id));
    }
}

/** Get Ufo by index.
    \param index [1,getNumUfos()] */
GUfo&
GUfoType::getUfoByIndex(int32_t index)
{
    // FIXME: this routine *must* be made smarter. This implementation effectively
    // makes all iteration over Ufos O(n).
    iterator it = begin();
    while (index > 1)
        ++it, --index;
    return *it;
}
