/**
  *  \file int/if/playerif.cc
  *  \brief Interpreter Game Interface: Player Properties
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/playerif.h"
#include "int/value.h"
#include "game/struct.h"
#include "game/racename.h"
#include "game/config.h"
#include "game/global.h"
#include "game/team.h"
#include "cpluslib/misc.h"
#include "int/if/ifutil.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/interror.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "int/propenum.h"
#include "int/contextio.h"

/** Implementation of "Player(nn)".
    \param args function parameters
    \return value representing "Player(nn)" */
IntValue*
IFPlayerGet(IntArgBlock& args)
{
    int32_t pid;
    args.checkArgumentCount(1);
    if (!checkIntArg(pid, args.getNext(), 1, NUM_PLAYERS))
        return 0;
    return new IntPlayerContext(pid);
}

/** Implementation of "ForEach Player".
    \return Context representing first player */
IntContext*
IFPlayerMake()
{
    return new IntPlayerContext(1);
}

/**************************** IntPlayerContext ***************************/

namespace {
    enum { PlayerProperty };
}

static const TIntMapping player_mapping[] = {
    // FIXME: missing PBPs
    { "BASES",               iplScoreBases,       PlayerProperty, thInt },
    { "PLANETS",             iplScorePlanets,     PlayerProperty, thInt },
    { "RACE",                iplFullName,         PlayerProperty, thString },
    { "RACE$",               iplId,               PlayerProperty, thInt },
    { "RACE.ADJ",            iplAdjName,          PlayerProperty, thString },
    { "RACE.ID",             iplRaceId,           PlayerProperty, thInt },
    { "RACE.MISSION",        iplMission,          PlayerProperty, thInt },
    { "RACE.SHORT",          iplShortName,        PlayerProperty, thString },
    { "SCORE",               iplScore,            PlayerProperty, thInt },
    { "SHIPS",               iplScoreShips,       PlayerProperty, thInt },
    { "SHIPS.CAPITAL",       iplScoreCapital,     PlayerProperty, thInt },
    { "SHIPS.FREIGHTERS",    iplScoreFreighters,  PlayerProperty, thInt },
    { "TEAM",                iplTeam,             PlayerProperty, thInt },
};

IntPlayerContext::IntPlayerContext(int pid)
    : pid(pid)
{ }

IntPlayerContext::~IntPlayerContext()
{ }

int32_t
IntPlayerContext::lookup(const string_t& name)
{
    return lookupName(name, player_mapping, countof(player_mapping));
}

void
IntPlayerContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntPlayerContext::get(int32_t index)
{
    return getPlayerProperty(pid, IntPlayerProperty(player_mapping[index].index));
}

bool
IntPlayerContext::next()
{
    if (pid < NUM_PLAYERS) {
        ++pid;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntPlayerContext::clone()
{
    return new IntPlayerContext(pid);
}

GObject*
IntPlayerContext::getObject()
{
    return 0;
}

string_t
IntPlayerContext::toString(bool /*readable*/)
{
    return format("Player(%d)", pid);
}

void
IntPlayerContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Player;
    sv.value = pid;
}

void
IntPlayerContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(player_mapping, countof(player_mapping));
}

static int
sumFreighters(const TGen& g)
{
    int sum = 0;
    for (int i = 0; i < NUM_PLAYERS; ++i)
        sum += g.scores[i].num_freighters;
    return sum;
}

static int
sumCapital(const TGen& g)
{
    int sum = 0;
    for (int i = 0; i < NUM_PLAYERS; ++i)
        sum += g.scores[i].num_capital;
    return sum;
}

/** Get player property.
    \param pid Player to query (out of range values are handled properly)
    \param ipl Player property to query */
IntValue*
getPlayerProperty(int pid, IntPlayerProperty ipl)
{
    /* Special case for out-of-range values */
    if (pid <= 0 || pid > NUM_PLAYERS || !haveDisplayedTurn()) {
        if (ipl == iplId && pid >= 0)
            return makeIntValue(pid);
        else
            return 0;
    }

    /* Regular handling */
    const TGen& gen = getDisplayedTurn().getGen(::getPlayerId()).getGenRecord();
    switch (ipl) {
     case iplAdjName:
        /*+ Adjective name of this player. */
        return makeStringValue(player_racenames.getAdjName(pid));
     case iplFullName:
        /*+ Full name of this player. */
        return makeStringValue(player_racenames.getFullName(pid));
     case iplId:
        /*+ Player number. */
        return makeIntValue(pid);
     case iplMission:
        /*+ Special mission assigned to this player. */
        return makeIntValue(raceMissionId(pid));
     case iplRaceId:
        /*+ Race assigned to this player. */
        return makeIntValue(raceId(pid));
     case iplShortName:
        /*+ Short name of this player. */
        return makeStringValue(player_racenames.getShortName(pid));
     case iplTeam:
        /*+ Team this player is in. */
        return makeIntValue(getPlayerTeam(pid));
     case iplScoreBases:
        /*+ Number of bases this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_bases);
     case iplScorePlanets:
        /*+ Number of planets this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_planets);
     case iplScoreShips:
        /*+ Number of ships this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_capital + gen.scores[pid-1].num_freighters);
     case iplScore:
        /*+ This player's Tim-score. */
        return makeIntValue(120L*gen.scores[pid-1].num_bases
                            + 10L*(gen.scores[pid-1].num_capital + gen.scores[pid-1].num_planets)
                            + gen.scores[pid-1].num_freighters);
     case iplScoreCapital:
        /*+ Number of capital ships this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_capital);
     case iplScoreFreighters:
        /*+ Number of freighters this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_freighters);
     case iplTotalShips:
        /*+ Total number of ships (from this player's score information). */
        return makeIntValue(sumFreighters(gen) + sumCapital(gen));
     case iplTotalFreighters:
        /*+ Total number of freighters (from this player's score information). */
        return makeIntValue(sumFreighters(gen));
     case iplTotalCapital:
        /*+ Total number of capital ships (from this player's score information). */
        return makeIntValue(sumCapital(gen));
    }
    return 0;
}
