/**
  *  \file int/if/shipif.cc
  *  \brief Interpreter Game Interface: Ships
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/shipif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/actions/basefix.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/objtypes.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"        // getHeadingName
#include "int/args.h"
#include "int/if/hullif.h"
#include "int/if/ifutil.h"
#include "int/if/objif.h"
#include "int/if/playerif.h"
#include "int/if/specif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "game/cspec.h"
#include "int/contextio.h"
#include "int/exec.h"
#include "int/process.h"
#include "int/propenum.h"
#include "game/cargotr.h"
#include "game/fleet.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "game/shippredict.h"

/****************************** IntShipValue *****************************/

/** Implementation of "Ship(nn)".
    \param args function parameters
    \return value representing "Ship(nn)" */
IntValue*
IFShipGet(IntArgBlock& args)
{
    int32_t id;

    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_SHIPS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntShipContext(id);
}

/** Implementation of "ForEach Ship".
    \return Context representing first ship */
IntContext*
IFShipMake()
{
    IntShipContext* sc = new IntShipContext(0);
    return sc->initFirst();
}

/****************************** Ship Methods *****************************/

static void
IFShipSetComment(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (IntValue* v = args.getNext()) {
        IntStringValue sv(v->toString(false));
        int sid = obj.getId();
        getShipProperties(sid).set(sp_Comment, &sv);
        univ.ty_any_ships.getObjectByIndex(sid).markDirty();
    }
}

static void
IFShipSetFCode(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispFCode, args.getNext());
}

static void
IFShipSetEnemy(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispEnemyId, args.getNext());
}

static void
IFShipSetSpeed(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispSpeedId, args.getNext());
}

static void
IFShipSetName(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispName, args.getNext());
}

static void
IFShipSetMission(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t m;
    int32_t i = 0, t = 0;

    // Evaluate arguments
    args.checkArgumentCount(1, 3);
    if (!checkIntArg(m, args.getNext(), 0, 10000))
        return;
    checkIntArg(i, args.getNext(), 0, 10000);
    checkIntArg(t, args.getNext(), 0, 10000);

    // Set mission on ship
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getShipKind() == GShip::CurrentShip) {
        if (!setFleetMission(univ, sh.getId(), m, i, t)) {
            throw GError(GError::eFleet, GError::eFleet);
        }
    } else {
        throw IntError::notAssignable();
    }
}

static void
setBaseOrder(GUniverse& univ, TShipyardAction order, GShip& sh)
{
    // Find the base we're at and verify the command
    int pid = univ.getPlanetAt(sh.getPos());
    if (GError e = checkBaseAction(order, univ, sh.getId(), pid))
        throw e;

    // Do it
    univ.getPlanet(pid).setBaseShipyardOrder(mp16_t(order), mp16_t(sh.getId()));
    sh.markDirty();
}

static void
IFShipFixShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(univ, sya_Fix, static_cast<GShip&>(obj));
}

static void
IFShipRecycleShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(univ, sya_Recycle, static_cast<GShip&>(obj));
}

static void
IFShipSetWaypoint(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t x, y;

    // Evaluate arguments
    args.checkArgumentCount(2);
    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;

    // Set ship waypoint
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getShipKind() == GShip::CurrentShip) {
        if (!setFleetWaypoint(univ, sh.getId(), GPoint(x, y)))
            throw GError(GError::eFleet, GError::eFleet);
    } else {
        throw IntError::notAssignable();
    }
}

static GCargoContainer*
makeShipTransfer(GShip& from, GShip& to)
{
    // Make transfer object for the "from" side
    if (from.getShipKind() == GShip::CurrentShip) {
        return new GShipTransfer(from);
    } else {
        if (to.isTransporterInUse(GShip::t_Transfer)) {
            throw IntError(GError::eFacility);
        }
        if (to.getTransporter(GShip::t_Transfer).id != 0
            && to.getTransporter(GShip::t_Transfer).id != from.getId()
            && to.isTransporterActive(GShip::t_Transfer))
        {
            if (!to.isTransporterCancellable(GShip::t_Transfer)) {
                throw IntError("Transporter jammed");
            }
            to.cancelTranporter(GShip::t_Transfer);
        }
        return new GShipTransporterTransfer(to, GShip::t_Transfer, from);
    }
}

GCargoContainer*
makePlanetTransfer(GUniverse& univ, int pid, GShip& to)
{
    if (pid == 0) {
        if (to.isTransporterInUse(GShip::t_Unload)) {
            throw IntError(GError::eFacility);
        }
        return new GShipTransporterTransfer(to);
    } else {
        GPlanet& pl = univ.getPlanet(pid);
        if (pl.isVisible() && pl.hasFullPlanetData() && pl.getOwner() == to.getOwner()) {
            // Direct transfer
            return new GPlanetTransfer(pl);
        } else {
            if (to.isTransporterInUse(GShip::t_Unload)) {
                throw IntError(GError::eFacility);
            }
            return new GShipTransporterTransfer(to, GShip::t_Unload, pl);
        }
    }
}

namespace {
    enum {
        flagOverload = 1,
        flagSup2MC   = 2,
        flagNoFail   = 4,
        flagJettison = 8
    };
}

/** Script-based cargo transfer.
    \param exc       Execution context of script
    \param transfer  Readily-configured GCargoTransfer
    \param cargospec Cargospec provided by user
    \param side      GCargoTransfer side to take away cargo from
    \param flags     Flags provided by user */
void
doScriptTransfer(IntExecutionContext& exc, GCargoTransfer& transfer, string_t cargospec, int side, int flags)
{
    // Parse cargospec
    GCargoSpec cs;
    if (!cs.parse(cargospec, false))
        throw IntError("Invalid cargo specification");

    // Configure
    if (flags & flagOverload) {
        transfer[0].setFlags(transfer[0].getFlags() | GCargoContainer::can_Overload);
        transfer[1].setFlags(transfer[1].getFlags() | GCargoContainer::can_Overload);
    }
    bool sup2mc = (flags & flagSup2MC) != 0
        && (transfer[0].hasFlag(GCargoContainer::can_Sup2mc)
            || transfer[1].hasFlag(GCargoContainer::can_Sup2mc));
    bool partial = (flags & flagNoFail) != 0;

    // Do it
    for (int i = 0; i <= el_Torps; ++i) {
        const GCargoType ct = GCargoType(i);
        if (cs.isValidType(ct)) {
            if (ct == el_Torps) {
                // This is a hack, but it's the same one as used in PCC 1.x :-)
                // Because GCargoTransfer needs an exact torpedo type, but we only
                // know that the user wants torpedoes, just try them all. Only one
                // of them will work (or none if the units are incompatible).
                for (int tt = 1; tt <= NUM_TORPS; ++tt)
                    cs.remove(ct, transfer.move(getCargoTypeFromTorpType(tt), cs.get(ct), side, partial, sup2mc));
            } else {
                cs.remove(ct, transfer.move(ct, cs.get(ct), side, partial, sup2mc));
            }
        }
    }

    // Evaluate
    if (partial) {
        IntStringValue sv(cs.toCargoSpecString());
        exc.setVariable("CARGO.REMAINDER", &sv);
    } else {
        if (!cs.isEmpty()) {
            throw IntError("Cargo transfer failed");
        }
    }
}

static void
IFShipCargoTransfer(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    // Args are cargospec, target ship Id, flags O/S/N (overload, supply sale, no-fail)
    string_t cargospec;
    int32_t target_sid;
    int32_t flags = 0;
    args.checkArgumentCount(2, 3);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    if (!checkIntArg(target_sid, args.getNext(), 1, NUM_SHIPS))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSN");

    // Validate
    GShip& sh       = static_cast<GShip&>(obj);
    GShip& target   = univ.getShip(target_sid);
    if (!target.isVisible() || !sh.isVisible()) {
        throw IntError("Ship does not exist");
    }
    if (target.getPos() != sh.getPos()) {
        throw IntError(GError::ePos);
    }
    if (target.getId() == sh.getId()) {
        // Forbid transfer from one ship to itself. PCC 1.x silently ignores this,
        // leading to subtle effects with Cargo.Remainder. So better fail it.
        // It's useless anyway.
        throw IntError(GError::ePerm);
    }
    if (target.getShipKind() != GShip::CurrentShip && sh.getShipKind() != GShip::CurrentShip) {
        // Neither is a current ship.
        throw IntError(GError::eNotOwner);
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(makeShipTransfer(sh, target));
    tx.addNewContainer(makeShipTransfer(target, sh));
    doScriptTransfer(exc, tx, cargospec, 0, flags);
    tx.doCommit();
}

static void
doShipCargoUnUpload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args, int side)
{
    // Args are cargospec, flags O/S/N/J (overload, supply sale, no-fail, jettison permission) + proxy Id
    // FIXME: proxy transfer support
    string_t cargospec;
    int32_t flags = 0;
    args.checkArgumentCount(1, 2);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSNJ");

    // Validate
    GShip& sh = static_cast<GShip&>(obj);
    if (!sh.isVisible()) {
        throw IntError("Ship does not exist");
    }
    if (sh.getShipKind() != GShip::CurrentShip) {
        throw IntError(GError::eNotOwner);
    }
    int pid = sh.getOrbitPlanetId();
    if (pid == 0 && (flags & flagJettison) == 0) {
        throw IntError("Not at planet");
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(new GShipTransfer(sh));
    tx.addNewContainer(makePlanetTransfer(univ, pid, sh));
    doScriptTransfer(exc, tx, cargospec, side, flags);
    tx.doCommit();
}

static void
IFShipCargoUnload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    doShipCargoUnUpload(obj, univ, exc, args, 0);
}

static void
IFShipCargoUpload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    doShipCargoUnUpload(obj, univ, exc, args, 1);
}

static void
IFShipSetFleet(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispFleetId, args.getNext());
}

/***************************** IntShipContext ****************************/

namespace {
    enum { ShipProperty, HullProperty, ComponentProperty, OwnerProperty, EnemyProperty, ShipMethod };
}

static const TIntMapping ship_mapping[] = {
    { "AUX",                       ispAuxName,                   ShipProperty,      thString },
    { "AUX$",                      ispAuxId,                     ShipProperty,      thInt },
    { "AUX.AMMO",                  ispAuxAmmo,                   ShipProperty,      thInt },
    { "AUX.COUNT",                 ispAuxCount,                  ShipProperty,      thInt },
    { "AUX.SHORT",                 ispAuxShort,                  ShipProperty,      thString },
    { "BEAM",                      ispBeamName,                  ShipProperty,      thString },
    { "BEAM$",                     ispBeamId,                    ShipProperty,      thInt },
    { "BEAM.COUNT",                ispBeamCount,                 ShipProperty,      thInt },
    { "BEAM.MAX",                  ihpMaxBeams,                  HullProperty,      thInt },
    { "BEAM.SHORT",                ispBeamShort,                 ShipProperty,      thString },
    { "CARGO.COLONISTS",           ispCargoColonists,            ShipProperty,      thInt },
    { "CARGO.D",                   ispCargoD,                    ShipProperty,      thInt },
    { "CARGO.FREE",                ispCargoFree,                 ShipProperty,      thInt },
    { "CARGO.M",                   ispCargoM,                    ShipProperty,      thInt },
    { "CARGO.MAX",                 ihpMaxCargo,                  HullProperty,      thInt },
    { "CARGO.MAXFUEL",             ihpMaxFuel,                   HullProperty,      thInt },
    { "CARGO.MONEY",               ispCargoMoney,                ShipProperty,      thInt },
    { "CARGO.N",                   ispCargoN,                    ShipProperty,      thInt },
    { "CARGO.STR",                 ispCargoStr,                  ShipProperty,      thString },
    { "CARGO.SUPPLIES",            ispCargoSupplies,             ShipProperty,      thInt },
    { "CARGO.T",                   ispCargoT,                    ShipProperty,      thInt },
    { "CARGOTRANSFER",             ismCargoTransfer,             ShipMethod,        thProcedure},
    { "CARGOUNLOAD",               ismCargoUnload,               ShipMethod,        thProcedure },
    { "CARGOUPLOAD",               ismCargoUpload,               ShipMethod,        thProcedure },
    { "CREW",                      ispCrew,                      ShipProperty,      thInt },
    { "CREW.NORMAL",               ihpMaxCrew,                   HullProperty,      thInt },
    { "DAMAGE",                    ispDamage,                    ShipProperty,      thInt },
    { "ENEMY",                     iplShortName,                 EnemyProperty,     thString },
    { "ENEMY$",                    ispEnemyId,                   ShipProperty,      thInt },
    { "ENEMY.ADJ",                 iplAdjName,                   EnemyProperty,     thString },
    { "ENGINE",                    ispEngineName,                ShipProperty,      thInt },
    { "ENGINE$",                   ispEngineId,                  ShipProperty,      thInt },
    { "ENGINE.COUNT",              ihpNumEngines,                HullProperty,      thInt },
    { "FCODE",                     ispFCode,                     ShipProperty,      thString },
    { "FIGHTER.BAYS",              ispFighterBays,               ShipProperty,      thInt },
    { "FIGHTER.COUNT",             ispFighterCount,              ShipProperty,      thInt },
    { "FIXSHIP",                   ismFixShip,                   ShipMethod,        thProcedure },
    { "FLEET",                     ispFleet,                     ShipProperty,      thString },
    { "FLEET$",                    ispFleetId,                   ShipProperty,      thInt },
    { "FLEET.NAME",                ispFleetName,                 ShipProperty,      thString },
    { "FLEET.STATUS",              ispFleetStatus,               ShipProperty,      thString },
    { "HEADING",                   ispHeadingName,               ShipProperty,      thString },
    { "HEADING$",                  ispHeadingAngle,              ShipProperty,      thInt },
    { "HULL",                      icpName,                      ComponentProperty, thString },
    { "HULL$",                     icpId,                        ComponentProperty, thInt },
    { "HULL.SHORT",                icpNameShort,                 ComponentProperty, thString },
    { "HULL.SPECIAL",              ispHullSpecial,               ShipProperty,      thString },
    { "ID",                        ispId,                        ShipProperty,      thInt },
    { "LEVEL",                     ispLevel,                     ShipProperty,      thInt },
    { "LOC",                       ispLoc,                       ShipProperty,      thString },
    { "LOC.X",                     ispLocX,                      ShipProperty,      thInt },
    { "LOC.Y",                     ispLocY,                      ShipProperty,      thInt },
    { "MARK",                      ismMark,                      ShipMethod,        thProcedure },
    { "MARKED",                    ispMarked,                    ShipProperty,      thBool },
    { "MASS",                      ispMass,                      ShipProperty,      thInt },
    { "MISSION",                   ispMissionName,               ShipProperty,      thString },
    { "MISSION$",                  ispMissionId,                 ShipProperty,      thInt },
    { "MISSION.INTERCEPT",         ispMissionIntercept,          ShipProperty,      thInt },
    { "MISSION.SHORT",             ispMissionShort,              ShipProperty,      thString },
    { "MISSION.TOW",               ispMissionTow,                ShipProperty,      thInt },
    { "MOVE.ETA",                  ispMoveETA,                   ShipProperty,      thInt },
    { "MOVE.FUEL",                 ispMoveFuel,                  ShipProperty,      thInt },
    { "NAME",                      ispName,                      ShipProperty,      thString },
    { "ORBIT",                     ispOrbitName,                 ShipProperty,      thString },
    { "ORBIT$",                    ispOrbitId,                   ShipProperty,      thInt },
    { "OWNER",                     iplShortName,                 OwnerProperty,     thString },
    { "OWNER$",                    iplId,                        OwnerProperty,     thInt },
    { "OWNER.ADJ",                 iplAdjName,                   OwnerProperty,     thString },
    { "OWNER.REAL",                ispRealOwner,                 ShipProperty,      thInt },
    { "PLAYED",                    ispPlayed,                    ShipProperty,      thBool },
    { "RECYCLESHIP",               ismRecycleShip,               ShipMethod,        thProcedure },
    { "SETCOMMENT",                ismSetComment,                ShipMethod,        thProcedure },
    { "SETENEMY",                  ismSetEnemy,                  ShipMethod,        thProcedure },
    { "SETFCODE",                  ismSetFCode,                  ShipMethod,        thProcedure },
    { "SETFLEET",                  ismSetFleet,                  ShipMethod,        thProcedure },
    { "SETMISSION",                ismSetMission,                ShipMethod,        thProcedure },
    { "SETNAME",                   ismSetName,                   ShipMethod,        thProcedure },
    { "SETSPEED",                  ismSetSpeed,                  ShipMethod,        thProcedure },
    { "SETWAYPOINT",               ismSetWaypoint,               ShipMethod,        thProcedure },
    { "SPEED",                     ispSpeedName,                 ShipProperty,      thString },
    { "SPEED$",                    ispSpeedId,                   ShipProperty,      thInt },
    { "TASK",                      ispTask,                      ShipProperty,      thBool },
    { "TECH.HULL",                 icpTech,                      ComponentProperty, thInt },
    { "TORP",                      ispTorpName,                  ShipProperty,      thString },
    { "TORP$",                     ispTorpId,                    ShipProperty,      thInt },
    { "TORP.COUNT",                ispTorpCount,                 ShipProperty,      thInt },
    { "TORP.LCOUNT",               ispTorpLCount,                ShipProperty,      thInt },
    { "TORP.LMAX",                 ihpMaxTorpLaunchers,          HullProperty,      thInt },
    { "TORP.SHORT",                ispTorpShort,                 ShipProperty,      thString },
    { "TRANSFER.SHIP",             ispTransferShip,              ShipProperty,      thBool },
    { "TRANSFER.SHIP.COLONISTS",   ispTransferShipColonists,     ShipProperty,      thInt },
    { "TRANSFER.SHIP.D",           ispTransferShipD,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.ID",          ispTransferShipId,            ShipProperty,      thInt },
    { "TRANSFER.SHIP.M",           ispTransferShipM,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.N",           ispTransferShipN,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.NAME",        ispTransferShipName,          ShipProperty,      thInt },
    { "TRANSFER.SHIP.SUPPLIES",    ispTransferShipSupplies,      ShipProperty,      thInt },
    { "TRANSFER.SHIP.T",           ispTransferShipT,             ShipProperty,      thInt },
    { "TRANSFER.UNLOAD",           ispTransferUnload,            ShipProperty,      thBool },
    { "TRANSFER.UNLOAD.COLONISTS", ispTransferUnloadColonists,   ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.D",         ispTransferUnloadD,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.ID",        ispTransferUnloadId,          ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.M",         ispTransferUnloadM,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.N",         ispTransferUnloadN,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.NAME",      ispTransferUnloadName,        ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.SUPPLIES",  ispTransferUnloadSupplies,    ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.T",         ispTransferUnloadT,           ShipProperty,      thInt },
    { "TYPE",                      ispTypeStr,                   ShipProperty,      thString },
    { "TYPE.SHORT",                ispTypeChar,                  ShipProperty,      thString },
    { "UNMARK",                    ismUnmark,                    ShipMethod,        thProcedure },
    { "WAYPOINT",                  ispWaypointName,              ShipProperty,      thString },
    { "WAYPOINT.DIST",             ispWaypointDistance,          ShipProperty,      thFloat },
    { "WAYPOINT.DX",               ispWaypointDX,                ShipProperty,      thInt },
    { "WAYPOINT.DY",               ispWaypointDY,                ShipProperty,      thInt },
    { "WAYPOINT.PLANET",           ispWaypointPlanetId,          ShipProperty,      thInt },
    { "WAYPOINT.X",                ispWaypointX,                 ShipProperty,      thInt },
    { "WAYPOINT.Y",                ispWaypointY,                 ShipProperty,      thInt },
};

static const IntObjectProcedureValue::call_t ship_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFShipSetComment,           // 2
    IFShipSetFCode,             // 3
    IFShipSetEnemy,             // 4
    IFShipSetSpeed,             // 5
    IFShipSetName,              // 6
    IFShipSetMission,           // 7
    IFShipFixShip,              // 8
    IFShipRecycleShip,          // 9
    IFShipSetWaypoint,          // 10
    IFShipCargoTransfer,        // 11
    IFShipCargoUnload,          // 12
    IFShipCargoUpload,          // 13
    IFShipSetFleet,             // 14
};

IntShipContext::IntShipContext(int sid)
    : IntObjectContext(current_anyship_type, sid)
{ }

IntShipContext::~IntShipContext()
{ }

static int32_t
lookupShipProperty(const string_t& name)
{
    static const int32_t nproperties = countof(ship_mapping);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = ship_property_names.lookup(name);
    if (ix != ship_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, ship_mapping, nproperties);
}

int32_t
IntShipContext::lookup(const string_t& name)
{
    /* Check prefix */
    if (name.compare(0, 5, "SHIP.", 5) == 0) {
        return lookupShipProperty(string_t(name, 5));
    } else {
        return lookupShipProperty(name);
    }
}

void
IntShipContext::set(int32_t index, IntValue* value)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip* sh = dynamic_cast<GShip*>(getObject());
        if (sh == 0) {
            throw IntError::notAssignable();
        }
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            setShipProperty(*sh, *getCurrentUniverse(), IntShipProperty(ship_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getShipProperties(getCurrentIndex()).set(index - countof(ship_mapping), value);
        getObject()->markDirty();
    }
}

IntValue*
IntShipContext::get(int32_t index)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip* sh = dynamic_cast<GShip*>(getObject());
        if (sh == 0) {
            return 0;
        }
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            return getShipProperty(*sh, *getCurrentUniverse(), IntShipProperty(ship_mapping[index].index));
         case HullProperty:
            if (sh->getHullId().isKnown())
                return getHullProperty(sh->getHull(), IntHullProperty(ship_mapping[index].index));
            else
                return 0;
         case ComponentProperty:
            if (sh->getHullId().isKnown())
                return getComponentProperty(sh->getHull(), IntComponentProperty(ship_mapping[index].index));
            else
                return 0;
         case OwnerProperty:
            return getPlayerProperty(sh->getOwner(), IntPlayerProperty(ship_mapping[index].index));
         case EnemyProperty:
            if (sh->getPrimaryEnemy().isKnown())
                return getPlayerProperty(sh->getPrimaryEnemy(), IntPlayerProperty(ship_mapping[index].index));
            else
                return 0;
         case ShipMethod:
            return new IntObjectProcedureValue(*sh, *getCurrentUniverse(), ship_methods[ship_mapping[index].index]);
         default:
            return 0;
        }
    } else {
        /* User property */
        return IntValue::cloneOf(getShipProperties(getCurrentIndex())[index - countof(ship_mapping)]);
    }
}

IntContext*
IntShipContext::clone()
{
    return new IntShipContext(getCurrentIndex());
}

string_t
IntShipContext::toString(bool /*readable*/)
{
    return format("Ship(%d)", getCurrentIndex());
}

void
IntShipContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Ship;
    sv.value = getCurrentIndex();
}

void
IntShipContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumNames(ship_property_names);
    acceptor.enumMapping(ship_mapping, countof(ship_mapping));
}

GObject*
IntShipContext::getObject()
{
    if (current_histship_type.isValidIndex(getCurrentIndex())) {
        return &current_histship_type.getObjectByIndex(getCurrentIndex());
    } else {
        return 0;
    }
}

/** Classify ship. This yields the ship's category as a string, 0 if unknown. */
static const char*
classifyShip(const GShip& sh)
{
    int beams = sh.getNumBeams().isKnown()         ? sh.getNumBeams()         : sh.getHullId().isKnown() ? sh.getHull().getMaxBeams()     : -1;
    int tubes = sh.getNumTorpLaunchers().isKnown() ? sh.getNumTorpLaunchers() : sh.getHullId().isKnown() ? sh.getHull().getMaxLaunchers() : -1;
    int bays  = sh.getNumBays().isKnown()          ? sh.getNumBays()          : sh.getHullId().isKnown() ? sh.getHull().getNumBays()      : -1;

    if (bays > 0)
        return "Carrier";
    else if (tubes > 0)
        return "Torpedo Ship";
    else if (beams > 0)
        return "Beam Weapons";
    else if (tubes == 0 && beams == 0 && bays == 0)
        return "Freighter";
    else
        return 0;
}

/** Get ship property.
    \param sh Ship to query
    \param isp Property to query */
IntValue*
getShipProperty(const GShip& sh, GUniverse& univ, const IntShipProperty isp)
{
    switch (isp) {
     case ispAuxId:
        /* 11 for fighters, torpedo tech if any, otherwise null */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeIntValue(11);
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeIntValue(sh.getTorpType());
        else
            return 0;
     case ispAuxAmmo:
        /* ammo */
        return makeIntFromMaybe(sh.getAmmoRaw());
     case ispAuxCount:
        /* number of aux weapons */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeIntValue(sh.getNumBays());
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0 && sh.getNumTorpLaunchers().isKnown())
            return makeIntValue(sh.getNumTorpLaunchers());
        else
            return 0;
     case ispAuxShort:
        /* short name of aux weapon */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeStringValue("Ftr");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispAuxName:
        /* full name of aux weapon */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeStringValue("Fighters");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispBeamId:
        /* beam type number */
        return makeIntFromMaybe(sh.getBeamType());
     case ispBeamCount:
        /* beam count */
        return makeIntFromMaybe(sh.getNumBeams());
     case ispBeamShort:
        /* short name of beam */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getShortName());
        else
            return 0;
     case ispBeamName:
        /* beam type */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getName());
        else
            return 0;
     case ispCargoColonists:
        /* colonists */
        return makeIntFromMaybe(sh.getCargoRaw(el_Colonists));
     case ispCargoD:
        /* duranium */
        return makeIntFromMaybe(sh.getCargoRaw(el_Duranium));
     case ispCargoFree:
        /* free cargo */
        return makeIntFromMaybe(sh.getFreeCargo());
     case ispCargoM:
        /* molybdenum */
        return makeIntFromMaybe(sh.getCargoRaw(el_Molybdenum));
     case ispCargoMoney:
        /* money */
        return makeIntFromMaybe(sh.getCargoRaw(el_Money));
     case ispCargoN:
        /* fuel */
        return makeIntFromMaybe(sh.getCargoRaw(el_Neutronium));
     case ispCargoStr:
        /* cargo string */
        {
            GCargoSpec sp;
            bool hadAny = false;
            for (GCargoType i = el_Neutronium; i <= el_Money; ++i) {
                mp16_t n = sh.getCargoRaw(i);
                if (n.isKnown()) {
                    sp.set(i, n);
                    hadAny = true;
                }
            }
            if (sh.getTorpType().isKnown() && sh.getTorpType() > 0) {
                mp16_t n = sh.getAmmoRaw();
                if (n.isKnown()) {
                    sp.set(el_Torps, n);
                }
            }

            if (hadAny) {
                return makeStringValue(sp.toCargoSpecString());
            } else {
                return 0;
            }
        }
     case ispCargoSupplies:
        /* supplies */
        return makeIntFromMaybe(sh.getCargoRaw(el_Supplies));
     case ispCargoT:
        /* supplies */
        return makeIntFromMaybe(sh.getCargoRaw(el_Tritanium));
     case ispCrew:
        /* crew */
        return makeIntFromMaybe(sh.getCrew());
     case ispDamage:
        /* damage */
        return makeIntFromMaybe(sh.getDamage());
     case ispEnemyId:
        /* enemy */
        return makeIntFromMaybe(sh.getPrimaryEnemy());
     case ispEngineId:
        /* engine id */
        return makeIntFromMaybe(sh.getEngineType());
     case ispEngineName:
        /* engine name */
        if (sh.getEngineType().isKnown())
            return makeStringValue(::getEngine(sh.getEngineType()).getName());
        else
            return 0;
     case ispFCode:
        /* fcode */
        if (sh.isFCodeKnown())
            return makeStringValue(sh.getFCode());
        else
            return 0;
     case ispFighterBays:
        /* fighter bays */
        return makeIntFromMaybe(sh.getNumBays());
     case ispFighterCount:
        /* fighter count */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0 && sh.getAmmoRaw().isKnown())
            return makeIntValue(sh.getAmmoRaw());
        else
            return 0;
     case ispFleetId:
        /* fleet Id */
        return makeIntValue(sh.getFleetNumber());
     case ispFleetName:
        /* fleet name */
        return makeStringValue(sh.getFleetName());
     case ispFleetStatus:
        /* membership status */
        if (int fid = sh.getFleetNumber()) {
            if (fid == sh.getId()) {
                return makeStringValue("leader");
            } else {
                return makeStringValue("member");
            }
        } else {
            return makeStringValue("-");
        }
     case ispFleet:
        /* fleet name */
        if (int fid = sh.getFleetNumber()) {
            GShip& leader = univ.getShip(fid);
            string_t result = leader.getFleetName();
            if (result.empty()) {
                result = leader.getName(GObject::LongName);
            }
            return makeStringValue(result);
        } else {
            return 0;
        }
     case ispHeadingAngle:
        /* heading angle */
        return makeIntFromMaybe(sh.getHeading());
     case ispHeadingName:
        /* heading name */
        // FIXME: this goes through NLS. Should it?
        if (sh.getHeading().isKnown())
            return makeStringValue(getHeadingName(sh.getHeading()));
        else
            return 0;
     case ispHullSpecial:
        /* Special flags */
        {
            string_t result;
            if (sh.isVisible()) {
                if (sh.canDoSpecial(hf_Cloak) || sh.canDoSpecial(hf_AdvancedCloak) || sh.canDoSpecial(hf_HardenedCloak))
                    result += "C";
                if (sh.canDoSpecial(hf_Hyperdrive))
                    result += "H";
                if (sh.canDoSpecial(hf_Gravitonic))
                    result += "G";
                if (sh.canDoSpecial(hf_Bioscan) || sh.canDoSpecial(hf_FullBioscan))
                    result += "B";
                if (sh.canDoSpecial(hf_MerlinAlchemy) || sh.canDoSpecial(hf_AriesRefinery) || sh.canDoSpecial(hf_NeutronicRefinery))
                    result += "A";
            }
            return makeStringValue(result);
        }
     case ispId:
        /* ship Id */
        return makeIntValue(sh.getId());
     case ispLevel:
        {
            int16_t level = sh.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ispLocX:
        /* ship location */
        if (sh.isVisible())
            return makeIntValue(sh.getPos().x);
        else
            return 0;
     case ispLocY:
        /* ship location */
        if (sh.isVisible())
            return makeIntValue(sh.getPos().y);
        else
            return 0;
     case ispLoc:
        if (sh.isVisible())
            return makeStringValue(univ.getLocationName(sh.getPos(), 0));
        else
            return 0;
     case ispMarked:
        /* marked */
        return makeBoolValue(sh.isMarked());
     case ispMass:
        /* mass */
        return makeIntFromMaybe(sh.getMass());
     case ispMissionId:
        /* mission */
        return makeIntFromMaybe(sh.getMission());
     case ispMissionIntercept:
        /* Intercept Id */
        return makeIntFromMaybe(sh.getInterceptId());
     case ispMissionShort:
        /* Short mission name */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByShip(sh);
            if (m != 0) {
                return makeStringValue(m->getShortName());
            } else {
                return makeStringValue(format("MIT %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMissionTow:
        /* Tow Id */
        return makeIntFromMaybe(sh.getTowId());
     case ispMissionName:
        /* mission name */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByShip(sh);
            if (m != 0) {
                return makeStringValue(m->getName());
            } else if (sh.getInterceptId().isKnown() && sh.getTowId().isKnown()) {
                return makeStringValue(format("M.I.T. %d (%d,%d)",
                                              int(sh.getMission()),
                                              int(sh.getInterceptId()),
                                              int(sh.getTowId())));
            } else {
                return makeStringValue(format("M.I.T. %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMoveETA:
        if (sh.getShipKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(univ, sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getNumTurns());
        } else {
            return 0;
        }
     case ispMoveFuel:
        if (sh.getShipKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(univ, sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getMovementFuelUsed());
        } else {
            return 0;
        }
     case ispName:
        if (sh.isVisible())
            return makeStringValue(sh.getName(GObject::PlainName));
        else
            return 0;
     case ispOrbitId:
        if (sh.isVisible())
            return makeIntValue(sh.getOrbitPlanetId());
        else
            return 0;
     case ispOrbitName:
        if (sh.getOrbitPlanetId() > 0)
            return makeStringValue(univ.getPlanet(sh.getOrbitPlanetId()).getName(GObject::PlainName));
        else
            return 0;
     case ispPlayed:
        return makeBoolValue(sh.isPlayed());
     case ispRealOwner:
        return makeIntValue(sh.getRealOwner());
     case ispSpeedId:
        return makeIntFromMaybe(sh.getWarp());
     case ispSpeedName:
        if (sh.getWarp().isKnown()) {
            if (sh.isHyperdriving())
                return makeStringValue("Hyperdrive");
            else
                return makeStringValue(format("Warp %d", sh.getWarp()));
        } else {
            return 0;
        }
     case ispTask:
        return makeBoolValue(getProcessByObject(&sh, IntExecutionContext::pkShipTask) != 0);
     case ispTorpId:
        return makeIntFromMaybe(sh.getTorpType());
     case ispTorpCount:
        if (sh.getTorpType().isKnown()) {
            if (sh.getTorpType() > 0)
                return makeIntFromMaybe(sh.getAmmo());
            else
                return makeIntValue(0);
        } else {
            return 0;
        }
     case ispTorpLCount:
        return makeIntFromMaybe(sh.getNumTorpLaunchers());
     case ispTorpShort:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispTorpName:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispTransferShipColonists:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).colonists);
        else
            return 0;
     case ispTransferShipD:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Duranium]);
        else
            return 0;
     case ispTransferShipId:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).id);
        else
            return 0;
     case ispTransferShipM:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferShipN:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferShipName:
        if (sh.getShipKind() == sh.CurrentShip
            && sh.isTransporterActive(sh.t_Transfer)
            && univ.ty_any_ships.isValidIndex(sh.getTransporter(sh.t_Transfer).id))
        {
            return makeStringValue(univ.getShip(sh.getTransporter(sh.t_Transfer).id).getName(GObject::PlainName));
        } else {
            return 0;
        }
     case ispTransferShipT:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferShipSupplies:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).supplies);
        else
            return 0;
     case ispTransferShip:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Transfer));
        else
            return 0;
     case ispTransferUnloadColonists:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).colonists);
        else
            return 0;
     case ispTransferUnloadD:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Duranium]);
        else
            return 0;
     case ispTransferUnloadId:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).id);
        else
            return 0;
     case ispTransferUnloadM:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferUnloadN:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferUnloadName:
        if (sh.getShipKind() == sh.CurrentShip && sh.isTransporterActive(sh.t_Unload)) {
            if (sh.getTransporter(sh.t_Unload).id == 0)
                return makeStringValue("Jettison");
            else if (univ.isValidPlanetId(sh.getTransporter(sh.t_Unload).id))
                return makeStringValue(univ.getPlanet(sh.getTransporter(sh.t_Unload).id).getName(GObject::PlainName));
            else
                return 0;
        } else {
            return 0;
        }
     case ispTransferUnloadT:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferUnloadSupplies:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).supplies);
        else
            return 0;
     case ispTransferUnload:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Unload));
        else
            return 0;
     case ispTypeChar:
        if (const char* p = classifyShip(sh))
            return makeStringValue(string_t(p, 1));
        else
            return 0;
     case ispTypeStr:
        if (const char* p = classifyShip(sh))
            return makeStringValue(p);
        else
            return 0;
     case ispWaypointDistance:
        if (sh.getShipKind() == sh.CurrentShip)
            return makeFloatValue(std::sqrt(double(square(sh.getWaypointDX()) + square(sh.getWaypointDY()))));
        else
            return 0;
     case ispWaypointDX:
        return makeIntFromMaybe(sh.getWaypointDX());
     case ispWaypointDY:
        return makeIntFromMaybe(sh.getWaypointDY());
     case ispWaypointPlanetId:
        if (sh.getShipKind() == sh.CurrentShip) {
            return makeIntValue(univ.getPlanetAt(sh.getWaypoint(), false));
        } else {
            return 0;
        }
     case ispWaypointX:
        return makeIntFromMaybe(mp16_t(sh.getWaypoint().x));
     case ispWaypointY:
        return makeIntFromMaybe(mp16_t(sh.getWaypoint().y));
     case ispWaypointName:
        if (sh.getShipKind() == sh.CurrentShip) {
            // FIXME: PCC 1.x also handles Intercept here
            if (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0)
                return makeStringValue("(Location)");
            else
                return makeStringValue(univ.getLocationName(sh.getWaypoint().getCanonicalLocation(), 0));
        } else {
            return 0;
        }
    }
    return 0;
}

/** Modify ship property.
    \param sh Ship to modify
    \param isp Property to modify
    \param value New value */
void
setShipProperty(GShip& sh, GUniverse& univ, const IntShipProperty isp, IntValue* value)
{
    string_t sv;
    int32_t iv;

    // Everything is only assignable for own ships. As an exception, the name is also assignable for targets.
    if (sh.getShipKind() != GShip::CurrentShip) {
        if (isp == ispName && sh.getShipKind() != GShip::NoShip) {
            // accept
        } else {
            throw IntError::notAssignable();
        }
    }

    switch (isp) {
     case ispFCode:
        if (checkStringArg(sv, value))
            sh.setFCode(sv);
        break;
     case ispMissionId:
     case ispMissionIntercept:
     case ispMissionTow:
        if (checkIntArg(iv, value, 0, 10000)) {
            // FIXME: what if one value is unknown?
            int m = (isp == ispMissionId        ? iv : sh.getMission());
            int i = (isp == ispMissionIntercept ? iv : sh.getInterceptId());
            int t = (isp == ispMissionTow       ? iv : sh.getTowId());
            if (!setFleetMission(univ, sh.getId(), m, i, t)) {
                throw GError(GError::eFleet, GError::eFleet);
            }
        }
        break;
     case ispName:
        if (checkStringArg(sv, value))
            sh.setName(sv);
        break;
     case ispSpeedId:
        if (checkIntArg(iv, value, 0, 9)) {
            if (!setFleetSpeed(univ, sh.getId(), iv)) {
                throw GError(GError::eFleet, GError::eFleet);
            }
        }
        break;
     case ispEnemyId:
        if (checkIntArg(iv, value, 0, NUM_PLAYERS))
            sh.setPrimaryEnemy(mp16_t(iv));
        break;
     case ispFleetName:
        if (checkStringArg(sv, value)) {
            if (!setFleetName(univ, sh.getId(), sv)) {
                throw IntError::notAssignable();
            }
        }
        break;
     case ispFleetId:
        if (checkIntArg(iv, value, 0, NUM_SHIPS)) {
            if (!setFleetNumber(univ, sh.getId(), iv)) {
                throw IntError::rangeError();
            }
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

void
callShipMethod(GShip& sh, GUniverse& univ, IntShipMethod ism, IntExecutionContext& exc, IntArgBlock& args)
{
    ship_methods[ism](sh, univ, exc, args);
}
